#!/usr/local/bin/perl

#	kana - test your knowledge of katakana
#
#	kana [option ...]

sub usage {
	die(
		"usage: kana [option ...]\n"
		. "\t-h\tshow hiragana\n"
		. "\t-k\tshow katakana (default)\n"
		. "\t-l <n>\ttest level n=0..9\n"
		. "\t-m\tinclude voiced kana\n"
		. "\t-c\tinclude compound kana\n"
		. "\t-s\tuse short format\n"
		. "\t-?\tprint this message\n"
	);
}

$zcat = 'zcat';
$gzcat = 'zcat';
$fontfile = '/usr/X11R6/lib/X11/fonts/misc/jiskan24.pcf';

# 21 -a  22 a   23 -i  24 i   25 -u   26 u   27 -e  28 e   29 -o  2a o
# 2b ka  2c ga  2d ki  2e gi  2f ku   30 gu  31 ke  32 ge  33 ko  34 go
# 35 sa  36 za  37 shi 38 ji  39 su   3a zu  3b se  3c ze  3d so  3e zo
# 3f ta  40 da  41 chi 42 ji  43 -tsu 44 tsu 45 zu  46 te  47 de  48 to  49 do
# 4a na  4b ni  4c nu  4d ne  4e no
# 4f ha  50 ba  51 pa  52 hi  53 bi   54 pi  55 fu  56 bu  57 pu
# 58 he  59 be  5a pe  5b ho  5c bo   5d po
# 5e ma  5f mi  60 mu  61 me  62 mo
# 63 -ya 64 ya  65 -yu 66 yu  67 -yo  68 yo
# 69 ra  6a ri  6b ru  6c re  6d ro
# 6e -wa 6f wa  70 wi  71 we  72 wo
# 73 n   74 wu  75 -ka 76 -ke

# 43 is the consonant doubler.

@allkana = (
	{"syl"=>"a", "code"=>0x22, "lev"=>0, "mod"=>0, "com"=>0},
	{"syl"=>"i", "code"=>0x24, "lev"=>0, "mod"=>0, "com"=>0},
	{"syl"=>"u", "code"=>0x26, "lev"=>0, "mod"=>0, "com"=>0},
	{"syl"=>"e", "code"=>0x28, "lev"=>0, "mod"=>0, "com"=>0},
	{"syl"=>"o", "code"=>0x2a, "lev"=>0, "mod"=>0, "com"=>0},

	{"syl"=>"ka", "code"=>0x2b, "lev"=>1, "mod"=>0, "com"=>0},
	{"syl"=>"ki", "code"=>0x2d, "lev"=>1, "mod"=>0, "com"=>0},
	{"syl"=>"ku", "code"=>0x2f, "lev"=>1, "mod"=>0, "com"=>0},
	{"syl"=>"ke", "code"=>0x31, "lev"=>1, "mod"=>0, "com"=>0},
	{"syl"=>"ko", "code"=>0x33, "lev"=>1, "mod"=>0, "com"=>0},

	{"syl"=>"sa", "code"=>0x35, "lev"=>2, "mod"=>0, "com"=>0},
	{"syl"=>"shi", "code"=>0x37, "lev"=>2, "mod"=>0, "com"=>0},
	{"syl"=>"su", "code"=>0x39, "lev"=>2, "mod"=>0, "com"=>0},
	{"syl"=>"se", "code"=>0x3b, "lev"=>2, "mod"=>0, "com"=>0},
	{"syl"=>"so", "code"=>0x3d, "lev"=>2, "mod"=>0, "com"=>0},

	{"syl"=>"ta", "code"=>0x3f, "lev"=>3, "mod"=>0, "com"=>0},
	{"syl"=>"chi", "code"=>0x41, "lev"=>3, "mod"=>0, "com"=>0},
	{"syl"=>"tsu", "code"=>0x44, "lev"=>3, "mod"=>0, "com"=>0},
	{"syl"=>"te", "code"=>0x46, "lev"=>3, "mod"=>0, "com"=>0},
	{"syl"=>"to", "code"=>0x48, "lev"=>3, "mod"=>0, "com"=>0},

	{"syl"=>"na", "code"=>0x4a, "lev"=>4, "mod"=>0, "com"=>0},
	{"syl"=>"ni", "code"=>0x4b, "lev"=>4, "mod"=>0, "com"=>0},
	{"syl"=>"nu", "code"=>0x4c, "lev"=>4, "mod"=>0, "com"=>0},
	{"syl"=>"ne", "code"=>0x4d, "lev"=>4, "mod"=>0, "com"=>0},
	{"syl"=>"no", "code"=>0x4e, "lev"=>4, "mod"=>0, "com"=>0},

	{"syl"=>"ha", "code"=>0x4f, "lev"=>5, "mod"=>0, "com"=>0},
	{"syl"=>"hi", "code"=>0x52, "lev"=>5, "mod"=>0, "com"=>0},
	{"syl"=>"fu", "code"=>0x55, "lev"=>5, "mod"=>0, "com"=>0},
	{"syl"=>"he", "code"=>0x58, "lev"=>5, "mod"=>0, "com"=>0},
	{"syl"=>"ho", "code"=>0x5b, "lev"=>5, "mod"=>0, "com"=>0},

	{"syl"=>"ma", "code"=>0x5e, "lev"=>6, "mod"=>0, "com"=>0},
	{"syl"=>"mi", "code"=>0x5f, "lev"=>6, "mod"=>0, "com"=>0},
	{"syl"=>"mu", "code"=>0x60, "lev"=>6, "mod"=>0, "com"=>0},
	{"syl"=>"me", "code"=>0x61, "lev"=>6, "mod"=>0, "com"=>0},
	{"syl"=>"mo", "code"=>0x62, "lev"=>6, "mod"=>0, "com"=>0},

	{"syl"=>"ya", "code"=>0x64, "lev"=>7, "mod"=>0, "com"=>0},
	{"syl"=>"yu", "code"=>0x66, "lev"=>7, "mod"=>0, "com"=>0},
	{"syl"=>"yo", "code"=>0x68, "lev"=>7, "mod"=>0, "com"=>0},

	{"syl"=>"ra", "code"=>0x69, "lev"=>8, "mod"=>0, "com"=>0},
	{"syl"=>"ri", "code"=>0x6a, "lev"=>8, "mod"=>0, "com"=>0},
	{"syl"=>"ru", "code"=>0x6b, "lev"=>8, "mod"=>0, "com"=>0},
	{"syl"=>"re", "code"=>0x6c, "lev"=>8, "mod"=>0, "com"=>0},
	{"syl"=>"ro", "code"=>0x6d, "lev"=>8, "mod"=>0, "com"=>0},

	{"syl"=>"wa", "code"=>0x6f, "lev"=>9, "mod"=>0, "com"=>0},
	{"syl"=>"(w)i", "code"=>0x70, "lev"=>9, "mod"=>0, "com"=>0},
	{"syl"=>"(w)e", "code"=>0x71, "lev"=>9, "mod"=>0, "com"=>0},
	{"syl"=>"(w)o", "code"=>0x72, "lev"=>9, "mod"=>0, "com"=>0},

# Omit "wu".  There's no historical hiragana for it, and the
# katakana belongs to the extended set, where it is called "vu".

#OMIT:	{"syl"=>"(w)u", "code"=>0x74, "lev"=>9, "mod"=>0, "com"=>0},

	{"syl"=>"n", "code"=>0x73, "lev"=>0, "mod"=>0, "com"=>0},

	{"syl"=>"ga", "code"=>0x2c, "lev"=>1, "mod"=>1, "com"=>0},
	{"syl"=>"gi", "code"=>0x2e, "lev"=>1, "mod"=>1, "com"=>0},
	{"syl"=>"gu", "code"=>0x30, "lev"=>1, "mod"=>1, "com"=>0},
	{"syl"=>"ge", "code"=>0x32, "lev"=>1, "mod"=>1, "com"=>0},
	{"syl"=>"go", "code"=>0x34, "lev"=>1, "mod"=>1, "com"=>0},

	{"syl"=>"za", "code"=>0x36, "lev"=>2, "mod"=>1, "com"=>0},
	{"syl"=>"ji", "code"=>0x38, "lev"=>2, "mod"=>1, "com"=>0},
	{"syl"=>"zu", "code"=>0x3a, "lev"=>2, "mod"=>1, "com"=>0},
	{"syl"=>"ze", "code"=>0x3c, "lev"=>2, "mod"=>1, "com"=>0},
	{"syl"=>"zo", "code"=>0x3e, "lev"=>2, "mod"=>1, "com"=>0},

	{"syl"=>"da", "code"=>0x40, "lev"=>3, "mod"=>1, "com"=>0},
	{"syl"=>"ji", "code"=>0x42, "lev"=>3, "mod"=>1, "com"=>0},
	{"syl"=>"zu", "code"=>0x45, "lev"=>3, "mod"=>1, "com"=>0},
	{"syl"=>"de", "code"=>0x47, "lev"=>3, "mod"=>1, "com"=>0},
	{"syl"=>"do", "code"=>0x49, "lev"=>3, "mod"=>1, "com"=>0},

	{"syl"=>"ba", "code"=>0x50, "lev"=>5, "mod"=>1, "com"=>0},
	{"syl"=>"pa", "code"=>0x51, "lev"=>5, "mod"=>1, "com"=>0},
	{"syl"=>"bi", "code"=>0x53, "lev"=>5, "mod"=>1, "com"=>0},
	{"syl"=>"pi", "code"=>0x54, "lev"=>5, "mod"=>1, "com"=>0},
	{"syl"=>"bu", "code"=>0x56, "lev"=>5, "mod"=>1, "com"=>0},
	{"syl"=>"pu", "code"=>0x57, "lev"=>5, "mod"=>1, "com"=>0},
	{"syl"=>"be", "code"=>0x59, "lev"=>5, "mod"=>1, "com"=>0},
	{"syl"=>"pe", "code"=>0x5a, "lev"=>5, "mod"=>1, "com"=>0},
	{"syl"=>"bo", "code"=>0x5c, "lev"=>5, "mod"=>1, "com"=>0},
	{"syl"=>"po", "code"=>0x5d, "lev"=>5, "mod"=>1, "com"=>0},

    {"syl"=>"kya", "code"=>0x2d, "code2"=>0x63, "lev"=>1, "mod"=>0, "com"=>1},
    {"syl"=>"kyu", "code"=>0x2d, "code2"=>0x65, "lev"=>1, "mod"=>0, "com"=>1},
    {"syl"=>"kyo", "code"=>0x2d, "code2"=>0x67, "lev"=>1, "mod"=>0, "com"=>1},

    {"syl"=>"sha", "code"=>0x37, "code2"=>0x63, "lev"=>2, "mod"=>0, "com"=>1},
    {"syl"=>"shu", "code"=>0x37, "code2"=>0x65, "lev"=>2, "mod"=>0, "com"=>1},
    {"syl"=>"sho", "code"=>0x37, "code2"=>0x67, "lev"=>2, "mod"=>0, "com"=>1},

    {"syl"=>"cha", "code"=>0x41, "code2"=>0x63, "lev"=>3, "mod"=>0, "com"=>1},
    {"syl"=>"chu", "code"=>0x41, "code2"=>0x65, "lev"=>3, "mod"=>0, "com"=>1},
    {"syl"=>"cho", "code"=>0x41, "code2"=>0x67, "lev"=>3, "mod"=>0, "com"=>1},

    {"syl"=>"nya", "code"=>0x4b, "code2"=>0x63, "lev"=>4, "mod"=>0, "com"=>1},
    {"syl"=>"nyu", "code"=>0x4b, "code2"=>0x65, "lev"=>4, "mod"=>0, "com"=>1},
    {"syl"=>"nyo", "code"=>0x4b, "code2"=>0x67, "lev"=>4, "mod"=>0, "com"=>1},

    {"syl"=>"hya", "code"=>0x52, "code2"=>0x63, "lev"=>5, "mod"=>0, "com"=>1},
    {"syl"=>"hyu", "code"=>0x52, "code2"=>0x65, "lev"=>5, "mod"=>0, "com"=>1},
    {"syl"=>"hyo", "code"=>0x52, "code2"=>0x67, "lev"=>5, "mod"=>0, "com"=>1},

    {"syl"=>"mya", "code"=>0x5f, "code2"=>0x63, "lev"=>6, "mod"=>0, "com"=>1},
    {"syl"=>"myu", "code"=>0x5f, "code2"=>0x65, "lev"=>6, "mod"=>0, "com"=>1},
    {"syl"=>"myo", "code"=>0x5f, "code2"=>0x67, "lev"=>6, "mod"=>0, "com"=>1},

    {"syl"=>"rya", "code"=>0x6a, "code2"=>0x63, "lev"=>8, "mod"=>0, "com"=>1},
    {"syl"=>"ryu", "code"=>0x6a, "code2"=>0x65, "lev"=>8, "mod"=>0, "com"=>1},
    {"syl"=>"ryo", "code"=>0x6a, "code2"=>0x67, "lev"=>8, "mod"=>0, "com"=>1},

    {"syl"=>"gya", "code"=>0x2e, "code2"=>0x63, "lev"=>1, "mod"=>1, "com"=>1},
    {"syl"=>"gyu", "code"=>0x2e, "code2"=>0x65, "lev"=>1, "mod"=>1, "com"=>1},
    {"syl"=>"gyo", "code"=>0x2e, "code2"=>0x67, "lev"=>1, "mod"=>1, "com"=>1},

    {"syl"=>"ja", "code"=>0x38, "code2"=>0x63, "lev"=>2, "mod"=>1, "com"=>1},
    {"syl"=>"ju", "code"=>0x38, "code2"=>0x65, "lev"=>2, "mod"=>1, "com"=>1},
    {"syl"=>"jo", "code"=>0x38, "code2"=>0x67, "lev"=>2, "mod"=>1, "com"=>1},

    {"syl"=>"bya", "code"=>0x53, "code2"=>0x63, "lev"=>5, "mod"=>1, "com"=>1},
    {"syl"=>"byu", "code"=>0x53, "code2"=>0x65, "lev"=>5, "mod"=>1, "com"=>1},
    {"syl"=>"byo", "code"=>0x53, "code2"=>0x67, "lev"=>5, "mod"=>1, "com"=>1},

    {"syl"=>"pya", "code"=>0x54, "code2"=>0x63, "lev"=>5, "mod"=>1, "com"=>1},
    {"syl"=>"pyu", "code"=>0x54, "code2"=>0x65, "lev"=>5, "mod"=>1, "com"=>1},
    {"syl"=>"pyo", "code"=>0x54, "code2"=>0x67, "lev"=>5, "mod"=>1, "com"=>1},
);

$level = 9;
$mflag = 0;
$cflag = 0;

while (@ARGV) {
	if ($ARGV[0] eq "-?") { &usage }
	elsif ($ARGV[0] eq "-l") {
		shift;
		@ARGV or &usage;
		$level = $ARGV[0];
	}
	elsif ($ARGV[0] =~ m{^-l}) {
		$level = $';
	}
	elsif ($ARGV[0] eq "-m") { $mflag = 1 }
	elsif ($ARGV[0] eq "-c") { $cflag = 1 }
	elsif ($ARGV[0] eq "-h") { $hflag = 1 }
	elsif ($ARGV[0] eq "-k") { $kflag = 1 }
	elsif ($ARGV[0] eq "-s") { $sflag = 1 }
#DEBUG#
	elsif ($ARGV[0] =~ m{^-d}) {
		$dflag = 1;
		$dvalue = $';
	}
#DEBUG#
	else { &usage }
	shift;
}

$level =~ m{^\d+$} or &usage;
$level <= 9 or &usage;

$hflag && $kflag and &usage;

$tempfile = "/tmp/kana.$$";
if (-e $fontfile) {
	system "cat $fontfile > $tempfile"
		and die "kana: cannot copy font file\n";
}
elsif (-e "$fontfile.Z") {
	system "$zcat $fontfile.Z > $tempfile"
		and die "kana: cannot uncompress font file\n";
}
elsif (-e "$fontfile.gz") {
	system "$gzcat $fontfile.gz > $tempfile"
		and die "kana: cannot uncompress font file\n";
}
else { die "kana: cannot find font file $fontfile\n" }

$SIG{INT} = "cleanup";
open FILE, "<$tempfile" or die "kana: cannot read $tempfile\n";
sub cleanup { unlink $tempfile; exit 0 }

tryread(FILE, \$header, 4);
# little-endian
$header eq "\1fcp" or die "kana: $fontfile is not a PCF file\n";

tryread(FILE, \$table_count, 4);
$table_count = unpack "V", $table_count;
for ($table=0; $table<$table_count; $table++) {
	tryread(FILE, \$desc, 16);
	($type, $format, $size, $offset) = unpack "VVVV", $desc;
	$type = whichbit($type);
	$byteorder = $format & (1<<2)? "MSBYTE": "LSBYTE";
	$bitorder = $format & (1<<3)? "MSBIT": "LSBIT";
	$comprshow = $type == 2 && ($format & (1<<8))? "compressed": "";
	$type==2 and $compr = $comprshow;
	($offset[$type], $size[$type], $byteorder[$type]) =
		($offset, $size, $byteorder eq "MSBYTE"? "N": "V");
}

$compr and $byteorder[2] = "c";		# compressed = signed bytes

&get_encoding;

seek FILE, $offset[3], 0 or die "pcffile: premature end of file\n";
tryread(FILE, \$format, 4);
$format = unpack "V", $format;
tryread(FILE, \$glyph_count, 4);
$glyph_count = unpack $byteorder[3], $glyph_count;

#DEBUG#
	if ($dflag) { &dump_glyph($dvalue); exit 0 }
#DEBUG#

# Select the specified kana from @allkana.
# Generate a random glyph from 25/21 to 25/75
#	(katakana 292..376, hiragana 209..291).

$basecode = $hflag? 209: 292;

@kana = grep {
		$_->{lev} <= $level
		&& (! $_->{mod} || $mflag)
		&& (! $_->{com} || $cflag)
	} @allkana;
warn((0+@kana) . " kana selected\n");

srand;
$prev = -1;
for (;;) {
	do { $new = int(@kana * rand) } until $new != $prev;
	$kana = $kana[$new];
	$code = $basecode + $kana->{code} - 0x21;
	$code2 = exists $kana->{code2}?
		$basecode + $kana->{code2} - 0x21: 0;
	dump_glyph($code, $code2);
	print "\nWhat is it? ";
	$junk = <STDIN>;	# TOKUM: check answer.
	$junk =~ m{^\s*q}i and last;	# quit.
	print $kana->{syl}, "\n";
	$prev = $new;
	sleep 1;
}

unlink $tempfile;
exit 0;

sub dump_glyph {
	my ($glnum, $glnum2) = @_;	# 2nd argument may be missing.
	local $byteorder = lc $byteorder[2];

	seek FILE, $offset[2], 0;
	tryread(FILE, \$format, 4);
	$format = unpack "V", $format;
	if ($compr) {
		$countsize = 2;
		$dsize = 5;
		tryread(FILE, \$metrics_count, $countsize);
		$metrics_count = unpack(($format & (1<<2)? "n": "v"),
			$metrics_count);
	}
	else {
		$countsize = 4;
		$dsize = 12;
		tryread(FILE, \$metrics_count, $countsize);
		$metrics_count = unpack(($format & (1<<2)? "N": "V"),
			$metrics_count);
	}

	$bitmap = read_glyph($glnum);
	$glnum2 and $bitmap2 = read_glyph($glnum2);

# Show the bitmap as a hex string:
#
#	$startc = "<";
#	$k = 0;
#	for ($j=0; $j<length $bitmap; $j++) {
#
#		# Skip the padding.
#
#		$j % $raster_width < (($character_width + 7)>>3) or next;
#		if (0 == $k++ % 32) {
#			print $startc;
#			$startc = "\n ";
#		}
#		printf "%02x", ord substr $bitmap, $j, 1;
#	}
#	print ">\n\n";

# Show the bitmap as a dot matrix:

	@glyphdisplay = ();
#	for ($y=0; $y<$character_height; $y++) {
	for ($y=0; $y<$character_height-2; $y++) {
		$raster = "";
		for ($x=0; $x<$character_width; $x++) {
			$ch = ord substr $bitmap,
				$raster_width * $y + ($x>>3), 1;
			$b = $bitorder eq "MSBIT"?
				0x80 >> ($x & 7):
				0x01 << ($x & 7);
			$raster .= (($b & $ch)? "*": " ");
		}
		if ($glnum2) {
			for ($x=0; $x<$character_width; $x++) {
				int $b;

				$ch = ord substr($bitmap2,
					$raster_width * $y + ($x>>3), 1);
				$b = $bitorder eq "MSBIT"?
					0x80 >> ($x & 7):
					0x01 << ($x & 7);
				$raster .= (($b & $ch)? "*": " ");
			}
		}
		push @glyphdisplay, $raster;
	}
	if ($sflag) {	# short format
		for (my $r = 0; $r < @glyphdisplay; $r += 2) {
			($hiraster, $loraster) = @glyphdisplay[$r..$r+1];
			for (my $c=0; $c<length $hiraster; $c++) {
				my $b = ((' ' ne substr $hiraster, $c, 1)<<1)
					|(' ' ne substr $loraster, $c, 1);
				print substr " .':", $b, 1;
			}
			print "\n";
		}
	}
	else { for (@glyphdisplay) { print "$_\n" } }
}

sub get_encoding {
	seek FILE, $offset[5], 0;
	tryread(FILE, \$eformat, 4);
	$eformat = unpack "V", $eformat;
	$eshort = $eformat & (1<<3)? "n": "v";
	tryread(FILE, \$string, 8);
	($low2, $high2, $low1, $high1) = unpack($eshort . 4, $string);
}

sub findcode {
	seek FILE, $offset[5] + 14 + $_[0] * 2, 0;
	tryread(FILE, \$index, 2);
	return unpack $eshort, $index;
}

sub read_glyph {
	my $glnum = shift;
	my $bm;

	seek FILE, $offset[2] + 4 + $countsize + $dsize * $glnum, 0;
	tryread(FILE, \$metrics, $dsize);

# Flip the sign bits.

	$compr and $metrics =~ s{.}{chr(0x80 ^ ord $&)}egs; # sneaky

	($left_sided_bearing, $right_sided_bearing,
	 $character_width, $character_ascent, $character_descent,
	 # $character_attributes
	) =
		unpack "${byteorder}5", $metrics;

	$character_height = $character_ascent + $character_descent;

# Get the bitmap.

	seek FILE, $offset[3]+8+$glnum*4, 0;
	tryread(FILE, \$boffset, 4);
	$boffset = unpack $byteorder[3], $boffset;
	$gboffset = $offset[3]+8+$glyph_count*4+16+$boffset;
	seek FILE, $gboffset, 0;

# Raster width in bytes = ((width + 31) & ~31) / 8;

	$raster_width = (($character_width + 31) & ~31) >> 3;

	tryread(FILE, \$bm, $raster_width * $character_height);
	return $bm;
}

sub tryread {
	my ($handle, $var, $len) = @_;
	my $result = read $handle, $$var, $len;
	$result == $len or die "pcffile: premature end of file\n";
}

sub whichbit {
	my $r = 0;
	until ($_[0] & 1) {
		$_[0] >>= 1;
		$r++;
	}
	return $r;
}
